# SPDX-License-Identifier: GPL-3.0-or-later
from __future__ import annotations

import copy
import math

import numpy as np
from numpy import complex128, float64
from numpy.typing import NDArray
from scipy.signal import welch

import deltares_wave_toolbox.cores.core_engine as core_engine
import deltares_wave_toolbox.series as series


def frequency_averaging(
    f: NDArray[float64],
    sVarDens: NDArray[float64],
    dfDesired: float = 0.0,
) -> tuple[NDArray[float64], NDArray[float64]]:
    """Band averaging of given variance density spectrum

    This function performs a band averaging on a given variance density
    spectrum sVarDens = sVarDens(f) on a frequency axis f onto a coarser
    frequency axis with frequency spacing dfDesired.

    Parameters
    ----------
    f : NDArray[float64]
        1D array containing frequency values, for folded Fourier transform. The number of elements in array f is
        close to half the number of elements in array xTime. To be precise, length(f) = floor(nT/2) + 1, with nT the
        number of elements in array xTime [Hz]
    sVarDens : NDArray[float64]
        1D array containing variance density spectrum of the signal [m^2/Hz]
    dfDesired : float, optional
        desired frequency spacing in Hertz on which the wave spectrum must be computed. If this parameter is omitted,
        then dfDesired = f(1) - f(0), by default 0.0 [Hz]

    Returns
    -------
    tuple[NDArray[float64], NDArray[float64]]
        f_coarse : NDArray[float64]
            frequency axis of computed spectrum. The frequency spacing is (close to) dfDesired [Hz]
        sVarDens_coarse : NDArray[float64]
            band averaged variance density spectrum of the signal on frequency axis f_coarse [m^2/Hz]

    Raises
    ------
    ValueError
        frequency_averaging: Input error: input should be 1d arrays

    Example
    -------
    >>> import numpy as np
    >>> df        = 0.01 #Hz
    >>> f         = np.arange(0,100)*df
    >>> sVarDens  = np.random.normal(loc=1, scale=2, size=len(f))
    >>> dfDesired = 0.02
    >>> f_coarse,sVarDens_coarse = frequency_averaging(f,sVarDens,dfDesired)

    """
    # convert input to array type to be able to handle input like e.g. f = [0.2,0.4]
    f, fSize = core_engine.convert_to_vector(f)
    sVarDens, sFreqSize = core_engine.convert_to_vector(sVarDens)

    if fSize[1] > 1 or sFreqSize[1] > 1:
        raise ValueError("frequency_averaging: Input error: input should be 1d arrays")

    # --- Determine some coefficients
    nFine = fSize[0]
    nFactor = 1
    if dfDesired != 0.0:
        nFactor = np.round(dfDesired / (f[1] - f[0]))

    # --- Avoid nFactor being equal to zero, which may occur if
    #     dfDesired < 0.5 * (f2 - f1)
    nFactor = max(nFactor, 1)
    nCoarse = math.floor(nFine / nFactor)

    # --- Initialize arrays
    f_coarse = np.zeros(len(f[0:nCoarse]))
    sVarDens_coarse = np.zeros(len(f_coarse), dtype=type(sVarDens[0]))

    # --- Perform the averaging
    for iFreq in np.arange(0, nCoarse):  # before np.arange(0,nCoarse)
        ilow = int((iFreq) * nFactor)
        ihigh = int((iFreq + 1) * nFactor)

        f_coarse[iFreq] = np.mean(f[ilow:ihigh])
        sVarDens_coarse[iFreq] = np.mean(sVarDens[ilow:ihigh])

    return f_coarse, sVarDens_coarse


def unfold_spectrum(
    f: NDArray[float64], xFreq: NDArray[complex128], isOdd: bool
) -> tuple[NDArray[float64], NDArray[complex128]]:
    """Unfolds a folded discrete Fourier transform

    This function unfolds a folded discrete Fourier transform xFreq that is given at frequency axis f. Note that this
    frequency axis goes up to the Nyquist frequency. Parameter isOdd indicates whether the underlying original time
    signal - of which xFreq is the discrete Fourier transform - has even (isOdd=0) or odd (isOdd=1) time points.

    The unfolded discrete Fourier xFreqTot, at frequency axis fTot, can be inverted back to time domain using
    function freq2time

    Parameters
    ----------
    f : NDArray[float64]
        1D array containing frequency values, for folded Fourier transform. The number of elements in array f is
        close to half the number of elements in array xTime. To be precise, length(f) = floor(nT/2) + 1, with nT the
        number of elements in array xTime [Hz]
    xFreq : NDArray[complex128]
        1D array (complex!) containing the folded Fourier coefficients of xTime. The value xFreq(i) must be the
        Fourier coefficient at frequency f(i). The number of elements in f and xFreq are the same.
    isOdd : bool
        logical indicating whether nT, the number of time points in original time series, is even (isOdd=False) or
        odd (isOdd=True)

    Returns
    -------
    tuple[NDArray[float64], NDArray[complex128]]
        fTot : NDArray[float64]
            1D array containing frequency values, for unfolded Fourier transform. The frequency axis runs from 0 to
            twice the Nyquist frequency. Array fTot contains as many elements as the original time series. [Hz]
        xFreqTot : NDArray[complex128]
            1D array (complex!) containing the unfolded Fourier coefficients of original time series. The value
            xFreqTot(i) must be the Fourier coefficient at frequency fTot(i). The number of elements in fTot and
            xFreqTot are the same.

    Raises
    ------
    ValueError
        Input error: input should be 1d arrays

    Example
    -------
    >>> import numpy as np
    >>> dt            = 0.1 # s
    >>> t             = np.arange(0,100) *dt  # Time axis
    >>> a1 = 0.5; w1  = 2*np.pi/5; phi1 = 0.35  # Wave component 1
    >>> a2 = 0.7; w2  = 2*np.pi/6; phi2 = 0.96  # Wave component 2
    >>> y = a1*np.cos(w1*t+phi1) + a2*np.cos(w2*t+phi2)
    >>> # --- Compute folded discrete Fourier transform
    >>> f,yFreq,isOdd = time2freq_nyquist(t,y);
    >>> # --- Unfold the discrete Fourier transform
    >>> fTot,yFreqTot = unfold_spectrum(f,yFreq,isOdd)
    >>> # --- Return to time domain
    >>> yTime         = freq2time(yFreqTot)  # yTime must be identical to y

    """
    # convert input to array type to be able to handle input like e.g. f = [0.2,0.4]
    f, fSize = core_engine.convert_to_vector(f)
    xFreq, xFreqSize = core_engine.convert_to_vector(xFreq)

    if fSize[1] > 1 or xFreqSize[1] > 1:
        raise ValueError("unfold_spectrum: Input error: input should be 1d arrays")

    nF = fSize[0]
    nFTot = (nF - 1) * 2 + isOdd

    # --- Construct frequency axis
    # Note that the first half part of the frequency axis is equal to the input
    # frequency axis as given in freq
    df = f[1] - f[0]
    fTot = np.arange(0, nFTot) * df

    xFreqTot = np.zeros((len(fTot)), dtype=complex)
    xFreqTot[0:nF] = copy.deepcopy(xFreq[0:nF])

    # Arrays f and xFreqTot are column vectors. Apply a flip upside-down
    xFreqTot[(nF - 1) + isOdd : nFTot] = np.flipud(np.conj(xFreqTot[1:nF]))

    return fTot, xFreqTot


def coherence(
    f: NDArray[float64],
    xFreq1: NDArray[complex128],
    xFreq2: NDArray[complex128],
    dfDesired: float = 0.0,
) -> tuple[NDArray[float64], NDArray[float64]]:
    """Function to compute the coherence in spectral domain.

    This function computes the coherence (magnitude-squared cohorence) of two complex spectral signal
    xFreq1 = xFreq1(f) and xFreq2 = xFreq2(f), given on frequency axis f. The output coh2 = coh2(f_coh2) is given on
    a frequency axis with frequency resolution dfDesired. Note: the coherence is real, and always between 0 and 1.

    Parameters
    ----------
    f : NDArray[float64]
        1D array containing frequency values, for folded Fourier transform. The number of elements in array f is
        close to half the number of elements in array xTime. To be precise, length(f) = floor(nT/2) + 1, with nT the
        number of elements in array xTime [Hz]
    xFreq1 : NDArray[complex128]
        1D array (complex!) containing the folded Fourier coefficients of signal 1. The value xFreq(i) must be the
        Fourier coefficient at frequency f(i). The number of elements in f and xFreq1 are the same.
    xFreq2 : NDArray[complex128]
        1D array (complex!) containing the folded Fourier coefficients of signal 2. The value xFreq(i) must be the
        Fourier coefficient at frequency f(i). The number of elements in f and xFreq2 are the same.
    dfDesired : float, optional
        desired frequency spacing in Hertz on which the wave spectrum must be computed. If this parameter is omitted,
        then dfDesired = f(1) - f(0), by default 0.0 [Hz]

    Returns
    -------
    tuple[NDArray[float64], NDArray[float64]]
        f_coh2 : NDArray[float64]
            frequency axis of coherence. The frequency spacing is (close to) dfDesired [Hz]
        coh2 : NDArray[float64]
            coherence (magnitude-squared coherence)
    """
    # --- Compute auto-spectral and cross-spectral (absolute value) densities
    #     on original frequency axis
    #     Note: Normalisation factors (2 / (df * Ntime * Ntime)) may be omitted,
    #     since they cancel out
    S11 = abs(xFreq1 * np.conj(xFreq1))
    S22 = abs(xFreq2 * np.conj(xFreq2))
    S12 = abs(xFreq1 * np.conj(xFreq2))

    # --- Apply band averaging
    f_coh2, S11_b = frequency_averaging(f, S11, dfDesired)
    f_coh2, S22_b = frequency_averaging(f, S22, dfDesired)
    f_coh2, S12_b = frequency_averaging(f, S12, dfDesired)

    # --- Omit division by zero, by putting the zeros (very small numbers) to a small
    #     number.
    small_number = 1e-10

    ismall = np.where(S11_b < small_number)[0]
    S11_b[ismall] = small_number

    ismall = np.where(S22_b < small_number)[0]
    S22_b[ismall] = small_number

    # --- Compute coherence
    coh2 = (S12_b) ** 2 / (S11_b * S22_b)

    return f_coh2, coh2


def freq2time(xFreq: NDArray[complex128]) -> NDArray[float64]:
    """Transforms (unfolded) discrete Fourier transform back to time signal

    This function transforms a given discrete and unfolded Fourier transform xFreq (in general a complex quantity!)
    back to time domain. Note that the input Fourier transform xFreq must be unfolded. A given folded Fourier
    transform can be unfolded using the function unfoldspectrum.

    Parameters
    ----------
    xFreq : NDArray[complex128]
        1D array (complex!) containing the folded Fourier coefficients of xTime. The value xFreq(i) must be the
        Fourier coefficient at frequency f(i). The number of elements in f and xFreq are the same.

    Returns
    -------
    NDArray[float64]
        1D array containing signal values, i.e. the time series of the signal. The value xTime(i) must be the signal
        value at time t(i). Usually water surface elevation [m]

    Raises
    ------
    ValueError
        Input error: input should be 1d arrays

    Example
    -------
    >>> import numpy as np
    >>> # --- Create time signal
    >>> dt =0.1   # s
    >>> t = np.arange(0,100)*dt  # Time axis
    >>> a1 = 0.5; w1 = 2*np.pi/5; phi1 = 0.35  # Wave component 1
    >>> a2 = 0.7; w2 = 2*np.pi/6; phi2 = 0.96  # Wave component 2
    >>> y = a1*np.cos(w1*t+phi1) + a2*np.cos(w2*t+phi2)
    >>> # --- Compute discrete Fourier transform
    >>> f,yFreq = time2freq(t,y)
    >>> # --- Return to time domain
    >>> yTime = freq2time(yFreq)  # yTime must be identical to y

    """
    xFreq, xFreqSize = core_engine.convert_to_vector(xFreq)

    # Check on input arguments
    #  xFreq must be a 1D array
    if xFreqSize[1] > 1:
        raise ValueError("freq2time: Input error: input should be 1d arrays")

    # Computational core
    nF = xFreqSize[0]
    xTime = np.real(np.fft.ifft(xFreq, nF))

    return xTime


def time2freq(
    t: NDArray[float64], xTime: NDArray[float64]
) -> tuple[NDArray[float64], NDArray[complex128]]:
    """Computes the discrete Fourier transform coefficients (on unfolded frequency axis) of a given of time signal

    This function computes the Fourier coefficients xFreq (in general complex quantities!) on frequency axis f
    (hence, xFreq = xFreq(f)), from a given time signal Xtime on time axis t (hence, xTime = xTime(t)).

    The Fourier transform is not folded. This means, that the number of elements in arrays f and xFreq is identical
    to nT (nT: the number of elements in arrays t and xTime). Note that the Fourier coefficients in xFreq have a
    complex conjugate symmetry around the Nyquist frequency. Transforming the signal xFreq = xFreq(f) back to time
    domain can be done with the function freq2time

    Parameters
    ----------
    t : NDArray[float64]
        1D real array containing time values. The numbers in the array t must be increasing and uniformly spaced
        (uniform time step). The initial time t(1) can be any value (so it is not obligatory to have t(1) = 0)
    xTime : NDArray[float64]
        1D array containing signal values, i.e. the time series of the signal. The value xTime(i) must be the signal
        value at time t(i). Usually water surface elevation [m]

    Returns
    -------
    tuple[NDArray[float64], NDArray[complex128]]
        f : NDArray[float64]
            1D array containing frequency values, for folded Fourier transform. The number of elements in array f is
            close to half the number of elements in array xTime. To be precise, length(f) = floor(nT/2) + 1, with nT
            the number of elements in array xTime [Hz]
        xFreq : NDArray[complex128]
            1D array (complex!) containing the unfolded Fourier coefficients of xTime. The value xFreq(i) must be the
            Fourier coefficient at frequency f(i). The number of elements in f and xFreq are the same. This number is
            the same as the number of elements in t and xTime.

    Raises
    ------
    ValueError
        Input error: input should be 1d arrays
    ValueError
        Input error: time input parameter must be monotonic with constant step size
    ValueError
        Input error: array sizes differ in dimension

    Example
    -------
    >>> import numpy as np
    >>> # --- Create time signal
    >>> dt =0.1 # s
    >>> t = np.arange(0,100)*dt  # Time axis
    >>> a1 = 0.5; w1 = 2*np.pi/5; phi1 = 0.35  # Wave component 1
    >>> a2 = 0.7; w2 = 2*np.pi/6; phi2 = 0.96  # Wave component 2
    >>> y = a1*np.cos(w1*t+phi1) + a2*np.cos(w2*t+phi2)
    >>> # --- Compute discrete Fourier transform
    >>> f,yFreq = time2freq(t,y)

    """
    t, tSize = core_engine.convert_to_vector(t)
    xTime, xTimeSize = core_engine.convert_to_vector(xTime)

    if tSize[1] > 1 or xTimeSize[1] > 1:
        raise ValueError("time2freq: Input error: input should be 1d arrays")

    if not core_engine.monotonic_increasing_constant_step(t):
        raise ValueError(
            "time2freq: Input error: time input parameter must be monotonic with constant step size"
        )

    if not (tSize[0] == xTimeSize[0]):
        raise ValueError("time2freq: Input error: array sizes differ in dimension")

    # Computational core
    # --- Initialize some constants
    #     Note: T = nT*dt = t(end) - t(1) + dt
    nT = tSize[0]
    dt = t[1] - t[0]
    T = nT * dt

    # --- Create frequency axis
    df = 1.0 / T
    f = np.arange(0, nT) * df  # exlcude f=0

    # --- Compute Fouriertransform
    xFreq = np.fft.fft(xTime)

    return f, xFreq


def time2freq_nyquist(
    t: NDArray[float64], xTime: NDArray[float64]
) -> tuple[NDArray[float64], NDArray[complex128], bool]:
    """Computes the discrete Fourier transform coefficients (on folded frequency axis) of a given of time signal

    This function computes the Fourier coefficients xFreq (in general complex quantities!) on frequency axis f
    (hence, xFreq = xFreq(f)), from a given time signal Xtime on time axis t (hence, xTime = xTime(t)).

    The Fourier transform is folded. This means, that the number of elements in arrays f and xFreq is identical to
    floor(nT/2) + 1 (nT: the number of elements in arrays t and xTime). This means that the the Fourier coefficients
    are computed up to the Nyquist frequency. Transforming the signal xFreq = xFreq(f) back to time domain can be
    done using first the function unfoldspectrum and after that the function freq2time. Parameter isOdd=1 if nT is
    odd, and isOdd=0 if nT is even.

    Parameters
    ----------
    t : NDArray[float64]
        1D array containing time axis. The numbers in the array t must be increasing and uniformly spaced
        (uniform time step) [s]
    xTime : NDArray[float64]
        1D array containing signal values, i.e. the time series of the signal. The value x(i) must be the signal
        value at time t(i). Usually water surface elevation [m]

    Returns
    -------
    tuple[NDArray[float64], NDArray[complex128], bool]
        f : NDArray[float64]
            1D array containing frequency values, for folded Fourier transform. The frequency axis runs from 0 to the
            Nyquist frequency. The number of elements in array f is close to half the number of elements in array
            xTime. To be precise, length(f) = floor(nT/2) + 1, with nT the number of elements in array xTime [Hz]
        xFreq : NDArray[complex128]
            1D array (complex!) containing the folded Fourier coefficients of xTime. The value xFreq(i) must be the
            Fourier coefficient at frequency f(i). The number of elements in f and xFreq are the same.
        isOdd : bool
            logical indicating whether nT, the number of time points in xTime, is even (isOdd=False) or odd
            (isOdd=True)

    Raises
    ------
    ValueError
        Input error: input should be 1d arrays
    ValueError
        Input error: time input parameter must be monotonic with constant step size
    ValueError
        Input error: array sizes differ in dimension

    Example
    -------
    >>> import numpy as np
    >>> dt        = 0.1 #Hz
    >>> t         = np.arange(0,100)*dt
    >>> a1 = 0.5; w1 = 2*np.pi/5; phi1 = 0.35;  # Wave component 1
    >>> a2 = 0.7; w2 = 2*np.pi/6; phi2 = 0.96;  # Wave component 2
    >>> y = a1*np.cos(w1*t+phi1) + a2*np.cos(w2*t+phi2)
    >>> print(t)
    >>> print(y)
    >>> # --- Compute discrete Fourier transform
    >>> f,yFreq,isOdd = time2freq_nyquist(t,y);

    """
    t, tSize = core_engine.convert_to_vector(t)
    xTime, xTimeSize = core_engine.convert_to_vector(xTime)

    if tSize[1] > 1 or xTimeSize[1] > 1:
        raise ValueError("time2freq_nyquist: Input error: input should be 1d arrays")

    if not core_engine.monotonic_increasing_constant_step(t):
        raise ValueError(
            "time2freq_nyquist: Input error: time input parameter must be monotonic with constant step size"
        )

    if not (tSize[0] == xTimeSize[0]):
        raise ValueError(
            "time2freq_nyquist: Input error: array sizes differ in dimension"
        )

    # Computational core Transform time signal to frequency domain, over
    # frequency axis up to twice the Nyquist frequency
    fTotal, xFreqTotal = time2freq(t, xTime)

    # --- Number of time points, and check whether this number is even or odd
    nT = tSize[0]
    isOdd = bool(nT % 2)

    # --- Index in array that corresponds to the Nyquist frequency
    #     Nyquist frequency = 1 / (2*dt)
    iNyq = math.floor(nT / 2) + 1

    # --- Take part of the signal up to the Nyquist frequency
    f = fTotal[0:iNyq]
    xFreq = xFreqTotal[0:iNyq]

    #
    return f, xFreq, isOdd


def compute_spectrum_time_series(
    t: NDArray[float64],
    xTime: NDArray[float64],
    dfDesired: float = 0.0,
) -> tuple[NDArray[float64], NDArray[float64]]:
    """Computes variance density spectrum from given time series

    This function computes a variance density spectrum sVarDens_coarse = sVarDens_coarse(f_coarse) on a frequency axis
    f_coarse from a given surface elevation time series xTime = xTime(t), with time axis t. The frequency spacing of
    the output spectrum is given by dfDesired.

    Parameters
    ----------
    t : NDArray[float64]
        1D array containing time axis. The numbers in the array t must be increasing and uniformly spaced
        (uniform time step) [s]
    xTime : NDArray[float64]
        1D array containing signal values, i.e. the time series of the signal. The value xTime(i) must be the signal
        value at time t(i). Usually water surface elevation [m]
    dfDesired : float, optional
        desired frequency spacing in Hertz on which the wave spectrum must be computed. If this parameter is omitted,
        then dfDesired = f(1) - f(0), by default 0.0 [Hz]

    Returns
    -------
    tuple[NDArray[float64], NDArray[float64]]
        f_coarse : NDArray[float64]
            frequency axis of computed spectrum. The frequency spacing is (close to) dfDesired [Hz]
        sVarDens_coarse : NDArray[float64]
            band averaged variance density spectrum of the signal on new frequency axis fCoarse [m^2/Hz]

    Example
    -------
    >>> import numpy as np
    >>> dt =0.1
    >>> t = np.arange(0,1000+dt,dt)  # Time axis
    >>> z = np.sin(t) + np.cos(2*t)  # Surface elevation data
    >>> df = 0.01                    # Choose value for frequency axis
    >>> [freq,varDens] = compute_spectrum_time_series(t,z,df)

    """
    # --- Transform to frequency domain ( input check is done in time2freq_nyquist)
    [f, xFreq, _] = time2freq_nyquist(t, xTime)
    df = f[1] - f[0]
    Ntime = len(t)
    sFine = 2 * xFreq * np.conj(xFreq) / (df * Ntime * Ntime)
    sFine = sFine.real

    # --- Perform averaging
    [f_coarse, sVarDens_coarse] = frequency_averaging(f, sFine, dfDesired)

    return f_coarse, sVarDens_coarse


def compute_spectrum_freq_series(
    f: NDArray[float64],
    xFreq: NDArray[float64],
    Ntime: int,
    dfDesired: float = 0.0,
) -> tuple[NDArray[float64], NDArray[float64]]:
    """Computes variance density spectrum from given complex spectrum of Fourier components

    This function computes a variance density spectrum sVarDens_coarse = sVarDens_coarse(f_coarse) on a frequency axis
    f_coarse from a given complex spectrum xFreq = xFreq(f) of Fourier coefficients on a frequency axis f. The
    frequency spacing of the output spectrum is given by dfDesired.

    Parameters
    ----------
    f : NDArray[float64]
        1D array containing frequency values, for folded Fourier transform. The number of elements in array f is
        close to half the number of elements in array xTime. To be precise, length(f) = floor(nT/2) + 1, with nT the
        number of elements in array xTime [Hz]
    xFreq : NDArray[float64]
        wave spectrum of complex Fourier coefficients
    dfDesired : float, optional
        desired frequency spacing in Hertz on which the wave spectrum must be computed. If this parameter is omitted,
        then dfDesired = f(1) - f(0), by default 0.0 [Hz]
    Ntime : int, optional
        number of time elements in original time signal, by default 0

    Returns
    -------
    tuple[NDArray[float64], NDArray[float64]]
        f_coarse : NDArray[float64]
            frequency axis of computed spectrum. The frequency spacing is (close to) dfDesired [Hz]
        sVarDens_coarse : NDArray[float64]
            band averaged variance density spectrum of the signal on new frequency axis fCoarse [m^2/Hz]

    Raises
    ------
    ValueError
        Input error: input should be 1d arrays
    ValueError
        Input error: frequency input parameter must be monotonic with constant step size
    ValueError
        Input error: array sizes differ in dimension

    """
    f, fSize = core_engine.convert_to_vector(f)
    xFreq, xFreqSize = core_engine.convert_to_vector(xFreq)

    if fSize[1] > 1 or xFreqSize[1] > 1:
        raise ValueError(
            "compute_spectrum_freq_series: Input error: input should be 1d arrays"
        )

    if not core_engine.monotonic_increasing_constant_step(f):
        raise ValueError(
            "compute_spectrum_freq_series: Input error: frequency input parameter must be monotonic with constant step"
            " size"
        )

    if not (fSize[0] == xFreqSize[0]):
        raise ValueError(
            "compute_spectrum_freq_series: Input error: array sizes differ in dimension"
        )

    # --- Transform to frequency domain
    df = f[1] - f[0]
    sFine = np.real(2 * xFreq * np.conj(xFreq) / (df * Ntime * Ntime))

    # --- Perform averaging
    [f_coarse, sVarDens_coarse] = frequency_averaging(f, sFine, dfDesired)

    return f_coarse, sVarDens_coarse


def spectrum2timeseries(
    f: NDArray[float64],
    sVarDens: NDArray[float64],
    tInit: float,
    tEnd: float,
    dt: float,
    seed: int = -1,
) -> tuple[NDArray[float64], NDArray[float64]]:
    """Generates a timeseries based on a given spectrum.

    This function synthesises a time series xTime = xTime(t) satisfying a given 1D variance density spectrum sVarDens
    = sVarDens(f). The time axis is specified by the user as well, starting at time tInit, ending at time tEnd, with
    a uniform time step equal to dt. The random phase method is used to synthetise the time series.

    Parameters
    ----------
    f : NDArray[float64]
        1D array containing frequency values, for folded Fourier transform. The number of elements in array f is
        close to half the number of elements in array xTime. To be precise, length(f) = floor(nT/2) + 1, with nT the
        number of elements in array xTime [Hz]
    sVarDens : NDArray[float64]
        1D array containing variance density spectrum of the signal [m^2/Hz]
    tInit : float
        initial time of time axis. [s]
    tEnd : float
        end time of time axis. [s]
    dt : float
        time step of time axis. [s]
    seed : int, optional
        Seed for random phases, by default -1

    Returns
    -------
    tuple[NDArray[float64], NDArray[float64]]
        t : NDArray[float64]
            time axis of computed spectrum. The time spacing is (close to) dt [s]
        xTime : NDArray[float64]
            time series of water level elevation [m]

    Raises
    ------
    ValueError
        Input error: input should be 1d arrays
    ValueError
        Input error: array sizes differ in dimension

    """
    f, fSize = core_engine.convert_to_vector(f)
    xFreq, xFreqSize = core_engine.convert_to_vector(sVarDens)

    if fSize[1] > 1 or xFreqSize[1] > 1:
        raise ValueError(
            "compute_spectrum_freq_series: Input error: input should be 1d arrays"
        )

    if not (fSize[0] == xFreqSize[0]):
        raise ValueError(
            "compute_spectrum_freq_series: Input error: array sizes differ in dimension"
        )

    t = np.arange(tInit, tEnd + dt, dt)

    nTime = len(t)
    nTimeEven = int(2 * np.ceil(nTime / 2))
    tDuration = nTimeEven * dt

    # --- Generate frequency axis fGen, with lowest frequency equal to df and
    #     highest frequency equal to the Nyquist frequency fNyq.
    #     Note that the the number of frequency points nF is equal to
    #     (nTimeEven / 2)
    dfGen = 1 / tDuration
    fNyq = 1 / (2 * dt)
    fGen = np.arange(dfGen, fNyq + dfGen / 2, dfGen)  # TODO why dfGen/2
    nF = len(fGen)

    # --- Interpolate the given variance density spectrum sVarDens = sVarDens(f)
    #     to a spectrum sVarDensGen = sVarDensGen(fGen), that is, at
    #     frequencies fGen.
    #     The values of sVarDensGen are put to zero outside the originally
    #     given range of f
    sVarDensGen = np.interp(fGen, f, sVarDens, left=0, right=0)

    # --- Chose nF random phases and, if requested, reset the state.
    #     Resetting the state means that a renewed call to the present function
    #     with the same arguments will give the same random numbers
    if seed >= 0:
        np.random.seed(seed)

    phase = 2 * np.pi * np.random.rand(1, nF)

    # --- Determine amplitudes per frequency band, based on the method of Miles
    amplitude = np.sqrt(2 * sVarDensGen * dfGen)

    # --- Construct the Fourier components

    xFreq = np.zeros(nTimeEven, dtype=complex128)
    xFreq[1 : nF + 1] = amplitude / 2.0 * np.exp(1j * phase)
    xFreq[nF:nTimeEven] = np.flip(np.conj(xFreq[1 : nF + 1]))

    # --- Inverse Fourier transformation to time domain
    xTime = np.real(np.fft.ifft(xFreq, nTimeEven) * nTimeEven)

    # --- Take the part corresponding to the time axis
    xTime = xTime[0:nTime]
    return t, xTime


def spectrum2timeseries_object(
    f: NDArray[float64],
    sVarDens: NDArray[float64],
    tInit: float,
    tEnd: float,
    dt: float,
    seed: int = -1,
) -> series.Series:
    """Generates a timeseries based on a given spectrum.

    This function synthesises a time series xTime = xTime(t) satisfying a given 1D variance density spectrum sVarDens
    = sVarDens(f). The time axis is specified by the user as well, starting at time tInit, ending at time tEnd, with
    a uniform time step equal to dt. The random phase method is used to synthetise the time series.

    Parameters
    ----------
    f : NDArray[float64]
        1D array containing frequency values, for folded Fourier transform. The number of elements in array f is
        close to half the number of elements in array xTime. To be precise, length(f) = floor(nT/2) + 1, with nT the
        number of elements in array xTime [Hz]
    sVarDens : NDArray[float64]
        1D array containing variance density spectrum of the signal [m^2/Hz]
    tInit : float
        initial time of time axis. [s]
    tEnd : float
        end time of time axis. [s]
    dt : float
        time step of time axis. [s]
    seed : int, optional
        Seed for random phases, by default -1

    Returns
    -------
    series.Series
        Series object containing the time series of water level elevation

    """
    t, xTime = spectrum2timeseries(f, sVarDens, tInit, tEnd, dt, seed)
    return series.Series(t, xTime)


def compute_spectrum_welch_wrapper(
    xTime: NDArray[float64],
    dt: float,
    nperseg: int = 0,
    noverlap: int = 0,
    nfft: int = 0,
    window_type: str = "hann",
) -> tuple[NDArray[float64], NDArray[float64]]:
    """Wrapper arround the Scipy Welch method

    Parameters
    ----------
    xTime : NDArray[float64]
        1D array containing signal values, i.e. the time series of the signal. The value xTime(i) must be the signal
        value at time t(i). Usually water surface elevation [m]
    dt : float
        time step of signal
    nperseg : int, optional
        Length of each segment, by default None
    noverlap : int, optional
        number of points in overlap, by default None
    nfft : int, optional
        length of fft, by default None
    window_type : str, optional
        window type, by default "hann"

    Returns
    -------
    tuple[NDArray[float64], NDArray[float64]]
        f : NDArray[float64]
            1D array containing frequency values, for folded Fourier transform. The number of elements in array f is
            close to half the number of elements in array xTime. To be precise, length(f) = floor(nT/2) + 1, with nT
            the number of elements in array xTime [Hz]
        sVarDens : NDArray[float64]
            1D array containing variance density spectrum of the signal [m^2/Hz]

    """
    if nperseg == 0:
        nperseg = None
    if noverlap == 0:
        noverlap = None
    if nfft == 0:
        nfft = None
    f, sVarDens = welch(
        xTime,
        fs=1 / dt,
        window=window_type,
        nperseg=nperseg,
        noverlap=noverlap,
        nfft=nfft,
        detrend="constant",
        return_onesided=True,
        scaling="density",
        axis=-1,
        average="mean",
    )
    return f, sVarDens


def applybandfilter(
    f: NDArray[float64], xFreq: NDArray[float64], flow: float, fhigh: float
):
    """applybandfilter

    Parameters
    ----------
    f : NDArray[float64]
        1D array containing the frequencies
    xFreq : NDArray[float64]
        1D array (complex!) containing the folded Fourier coefficients of xTime. The value xFreq(i) must be the
        Fourier coefficient at frequency f(i). The number of elements in f and xFreq are the same.
    flow : float
        Lower frequeny of the bandpassfilter
    fhigh : float
        Higher frequency of the bandpassfilter


    Returns
    -------
    tuple NDArray[float64]]
        zFreq : NDArray[float64]
            1D array containing the folded Fourier coefficients.

    """
    # --- Length of array f = length of array xFreq
    Nf = len(f)

    # --- Find index in frequency array corresponding to flow
    if flow <= f[0]:
        ilow = 1
    elif flow >= f[-1]:
        ilow = Nf
    else:
        ilow = np.where(f < flow)[0][-1] + 1

    # --- Find index in frequency array corresponding to fhigh
    if fhigh <= f[0]:
        ihigh = 1
    elif fhigh >= f[-1]:
        ihigh = Nf
    else:
        ihigh = np.where(f < fhigh)[0][-1]

    # --- Initialize
    zFreq = np.zeros(len(xFreq)).astype(complex)

    # --- Apply band filter
    zFreq[ilow:ihigh] = xFreq[ilow:ihigh]
    return zFreq


def bandpassfilter(
    t: NDArray[float64], xTime: NDArray[float64], flow: float, fhigh: float
):
    """Bandpassfilter
    Filters a signal based on a lower and higher frequency limit.

    Parameters
    ----------
    t : NDArray[float64]
        1D array containing the time
    xTime : NDArray[float64]
        1D array containing signal values, i.e. the time series of the signal. The value xTime(i) must be the signal
        value at time t(i). Usually water surface elevation [m]
    flow : float
        Lower frequeny of the bandpassfilter
    fhigh : float
        Higher frequency of the bandpassfilter


    Returns
    -------
    tuple NDArray[float64]]
        zTime : NDArray[float64]
            1D array containing the filtered signal.

    """
    [f, xFreq, isOdd] = time2freq_nyquist(t, xTime)

    zFreq = applybandfilter(f, xFreq, flow, fhigh)
    [fUnfold, zFreqUnfold] = unfold_spectrum(f, zFreq, isOdd)
    zTime = freq2time(zFreqUnfold)
    return zTime



# SPDX-License-Identifier: GPL-3.0-or-later
import copy

import numpy as np
from numpy import float64
from numpy.typing import NDArray

import deltares_wave_toolbox.cores.core_dispersion as core_dispersion
import deltares_wave_toolbox.cores.core_spectral as core_spectral
import deltares_wave_toolbox.series as series


def decompose_linear_ZS(
    t: NDArray[float64],
    xTime: NDArray[float64],
    h: NDArray[float64],
    x_loc: NDArray[float64],
    w_loc: NDArray[float64],
    detLim: float = 0.125,
) -> tuple[NDArray[float64], NDArray[float64]]:
    """
    decompose_linear_ZS  Separation of incident and reflected waves using Zelt & Skjelbreia's method.

    This function performs the separation of incident and reflected waves
    according to the Zelt & Skjelbreia's procedure. Time series of wave
    signals xTime(Ntime,Nloc) at locations x_loc(Nloc) need to be provided.
    All wave gauges are assigned a weight w_loc(Nloc).
    Waves can only propagate in the positive x-direction (incoming wave) and
    the negative x-direction (reflected wave). The water depth h is assumed
    to be uniform. Time series of the incident and reflected wave in
    location x_loc(1) are provided as output of this function.

    For Nloc = 2 and w_loc = [1 1], this reduces to Goda's and Suzuki's
    method.
    For Nloc = 3 and w_loc = [1 1 1], this reduces to Mansard-Funke method.

    Parameters
    ----------
    t      :
           1D real array containing Ntime time values. The numbers in the
           array t must be increasing and uniformly spaced (uniform time
           step). The initial time t(1) can be any value (so it is not
           obligatory to have t(1) = 0). Units: seconds
    xTime  :
           real array of size [Ntime,Nloc] containing the time series of
           the surface elevation in locations x_loc.
           Units: meters
    h      :
           depth, in meters. Assumed to be uniform
    x_loc  :
           1D real array with Nloc entries indicating the locations where
           the time signals are measured.
    w_loc  :
           1D real array with Nloc entries indicating the weights of the
           signals. The values of w_loc must be between zero and 1.
    detLim :
           Optional parameter, with default value 0.125.
           This parameter, which is essentially borrowed from the
           Mansard-Funke function, indicates the value of the matrix
           determinantabove which the Mansard-Funke procedure is applied.
           When thedeterminant is below the value detLim, a more crude
           splitting procedure is applied. The value of detLim must be
           small and positive. Parameter detLim is related to the
           parameter denominator in the Auke-Process input. The relation
           reads:
               denominator = 9 - 4 * detLim.
           Hence, for detLim = 0.125, we get denominator = 8.5 (default
           value in Auke-Process).

    Returns
    -------
    xTime1In :
             1D real array containing time signal of incident wave (i.e.
             propagating in positive x-direction), in location x_loc(1)
    xTime1Re :
             1D real array containing time signal of reflected wave (i.e.
             propagating in negative x-direction), in location x_loc(1)

    Remark:
       * The wave signals must be ordered such that x_loc(1) < x_loc(2) < ....
       * Nloc must be two or larger
       * The sum of the weights w_loc must be two or larger

    Syntax:
       xTime1In,xTime1Re = decompose_linear_ZS(t,xTime,h,x_loc,w_loc,detLim)

    """

    # --- Some checks on input

    # --- Check whether input arrays are consistent in the sense that they
    #     share the same number of locations
    Ntime = 0
    Nloc = 0
    Ntime, Nloc = xTime.shape
    Nloc2 = len(x_loc)
    Nloc3 = len(w_loc)

    assert (
        Nloc == Nloc2
    ), "ZS: input error: Dimensions of xTime and x_loc do not correspond."
    assert (
        Nloc2 == Nloc3
    ), "ZS: input error: Dimensions of x_loc and w_loc do not correspond."
    assert Nloc > 1, "ZS: input error: Number of wave gauges must be at least 2."

    # --- Check whether the number of time samples agrees
    assert Ntime == len(
        t
    ), "ZS: input error: Dimensions of x_loc and t do not correspond."
    # --- Values of w_loc should be between 0 and 1
    assert (
        min(w_loc) > 0 or max(w_loc) <= 1
    ), "ZS: input error: Values of w_loc must be between 0 and 1."

    w_loc_sum = sum(w_loc)
    assert (
        w_loc_sum > 1
    ), "ZS: input error: Sum of elements in w_loc must be at least 2."

    # --- Check whether x_loc(1) < x_loc(2) < ...
    assert (
        min(np.diff(x_loc)) > 0
    ), 'ZS: input error": Wave gauges positions x_loc are not sorted in increasing order.'

    # --- Computational core
    # --- Some constants
    g = 9.81

    #
    diff_x_loc = np.zeros(Nloc)

    for iloc in np.arange(1, Nloc):
        diff_x_loc[iloc] = x_loc[iloc] - x_loc[0]

    # --- Compute Fourier transform of the time signals
    f, xFreq1, isOdd = core_spectral.time2freq_nyquist(t, xTime[:, 0])
    Nf = len(f)

    # to solve waring: ComplexWarning: Casting complex values to real discards the imaginary part.
    # note 2d dimension is specified as one argument (Nf,Nloc) for function zeros()
    xFreq = np.zeros((Nf, Nloc), dtype=complex)
    xFreq[:, 0] = copy.deepcopy(xFreq1)
    # note iloc runs from 1 to Nloc-1 when using range(1,Nloc)
    for iloc in np.arange(1, Nloc):
        f, xFreq[:, iloc], isOdd = core_spectral.time2freq_nyquist(t, xTime[:, iloc])

    # --- Compute the wave number
    w = 2 * np.pi * f
    k = core_dispersion.disper(w, h, g)

    # --- Initialize complex spectrum of incident and reflected waves
    #     Their initial value is, as is done also in Auke-Process, set to zero.
    #     For most frequencies (i.e. the ones corresponding to a matrix
    #     determinant larger than detLim), the values as computed from the
    #     Mansard-Funke procedure are inserted in xFreq1In and xFreq1Re. At
    #     the frequencies with a determinant smaller than detLim (these are the
    #     smallest frequencies), the zero initialisation means that all energy
    #     at these frequencies is discarded. In the future perhaps a better
    #     initialisation approach must be found, such as for example:
    #        xFreq1In = (1-r)*xFreq1; xFreq1Re = r*xFreq1;
    #     where r is some user-defined parameter or a measure for the
    #     reflection.
    xFreq1In = np.zeros(len(xFreq1), dtype=complex)
    xFreq1Re = np.zeros(len(xFreq1), dtype=complex)

    # --- Loop over all frequencies
    a12 = w_loc_sum
    a21 = w_loc_sum
    for ifreq in np.arange(0, Nf):
        # --- Initialize matrix elements and right-hand sides
        a11 = 0
        a22 = 0
        b1 = 0
        b2 = 0
        # --- Compute matrix elements and right-hand sides
        for iloc in np.arange(0, Nloc):
            a11 = a11 + w_loc[iloc] * np.exp(-1j * 2 * k[ifreq] * diff_x_loc[iloc])
            a22 = a22 + w_loc[iloc] * np.exp(1j * 2 * k[ifreq] * diff_x_loc[iloc])
            #
            b1 = b1 + w_loc[iloc] * xFreq[ifreq, iloc] * np.exp(
                -1j * k[ifreq] * diff_x_loc[iloc]
            )
            b2 = b2 + w_loc[iloc] * xFreq[ifreq, iloc] * np.exp(
                1j * k[ifreq] * diff_x_loc[iloc]
            )

        # --- Compute determinant of matrix (for all frequency components)
        detA = a11 * a22 - a12 * a21

        # --- The scaled determinant sDetA is equal to:
        sDetA = abs(-0.25 * detA)

        # --- Consider only frequency components for which sDetA > detLim. For
        #     these frequency components, solve 2x2 system analytically
        if sDetA > detLim:
            xFreq1In[ifreq] = (a22 * b1 - a12 * b2) / detA
            xFreq1Re[ifreq] = (-a21 * b1 + a11 * b2) / detA

    # --- Unfold spectra
    fTot, xFreq1InTot = core_spectral.unfold_spectrum(
        f.reshape(len(f), 1), xFreq1In.reshape(len(f), 1), isOdd
    )
    fTot, xFreq1ReTot = core_spectral.unfold_spectrum(
        f.reshape(len(f), 1), xFreq1Re.reshape(len(f), 1), isOdd
    )

    # --- Transform back to time domain
    xTime1In = core_spectral.freq2time(xFreq1InTot)
    xTime1Re = core_spectral.freq2time(xFreq1ReTot)

    return xTime1In, xTime1Re


def decompose_linear_ZS_series(
    Series_objects: list[series.Series],
    h: NDArray[float64],
    x_loc: NDArray[float64],
    w_loc: NDArray[float64],
    detLim: float = 0.125,
):
    """Separation of incident and reflected waves using the Zelt & Skjelbreia (1992) method.

    This is a wrapper function for the decompose_linear_ZS function, made compatible with
    the Series class from the Deltares Wave Toolbox.

    This function performs the separation of incident and reflected waves
    according to the Zelt & Skjelbreia (1992) procedure. Time series of wave
    signals xTime(Ntime,Nloc) at locations x_loc(Nloc) need to be provided.
    All wave gauges are assigned a weight w_loc(Nloc).
    Waves can only propagate in the positive x-direction (incoming wave) and
    the negative x-direction (reflected wave). The water depth h is assumed
    to be uniform. Time series of the incident and reflected wave in
    location x_loc(1) are provided as output of this function.

    For Nloc = 2 and w_loc = [1 1], this reduces to Goda's and Suzuki's
    method.
    For Nloc = 3 and w_loc = [1 1 1], this reduces to Mansard-Funke method.

    Parameters
    ----------
    Series_objects : list[series.Series]
        List of Series objects containing the time series of the surface elevation in locations x_loc.
    h : NDArray[float64]
        depth, in meters. Assumed to be uniform
    x_loc : NDArray[float64]
        1D real array with Nloc entries indicating the locations where
        the time signals are measured.
    w_loc : NDArray[float64]
        1D real array with Nloc entries indicating the weights of the
        signals. The values of w_loc must be between zero and 1.
    detLim : float, optional
        Optional parameter, by default 0.125
           This parameter, which is essentially borrowed from the
           Mansard-Funke function, indicates the value of the matrix
           determinantabove which the Mansard-Funke procedure is applied.
           When thedeterminant is below the value detLim, a more crude
           splitting procedure is applied. The value of detLim must be
           small and positive. Parameter detLim is related to the
           parameter denominator in the Auke-Process input. The relation
           reads:
               denominator = 9 - 4 * detLim

    Returns
    -------
    xTime1In :
             1D real array containing time signal of incident wave (i.e.
             propagating in positive x-direction), in location x_loc(1)
    xTime1Re :
             1D real array containing time signal of reflected wave (i.e.
             propagating in negative x-direction), in location x_loc(1)
    """

    for i, series_object in enumerate(Series_objects):

        if i == 0:
            t = series_object.time
            xTime = np.zeros((len(t), len(x_loc)))

        xTime[:, i] = series_object.xTime

    xTime1In, xTime1Re = decompose_linear_ZS(t, xTime, h, x_loc, w_loc, detLim)

    xTime1In_series = series.Series(t, xTime1In)
    xTime1Re_series = series.Series(t, xTime1Re)

    return xTime1In_series, xTime1Re_series



# SPDX-License-Identifier: GPL-3.0-or-later
import numpy as np
from numpy import float64
from numpy.typing import NDArray

import deltares_wave_toolbox.cores.core_engine as core_engine


def sort_wave_params(
    hWave: NDArray[float64], tWave: NDArray[float64]
) -> tuple[NDArray[float64], NDArray[float64]]:
    """Sorts the wave height and wave period

    This functions sorts the wave heights and wave periods in arrays hWave and tWave, and stores them in arrays
    hWaveSorted and tWaveSorted. The sorting is done such that in hWaveSorted the wave heights of hWave are sorted in
    descending order. This same sorting is applied to tWave, with the result in tWaveSorted. This means that
    hWaveSorted(i) and hWaveSorted(i) correspond to the wave height and wave period of the same wave.

    Parameters
    ----------
    hWave : NDArray[float64]
        1D array containing the wave heights of the individual waves [m]
    tWave : NDArray[float64]
        1D array containing the periods of the individual waves [s]

    Returns
    -------
    tuple[NDArray[float64], NDArray[float64]]
        hWaveSorted : NDArray[float64]
            1D array containing wave heights of the individual waves, sorted in descending order [m]
        tWaveSorted : NDArray[float64]
            1D array containing wave periods of the individual waves, using the same sorting (re-arranging) as
            applied to hWave [s]

    Raises
    ------
    ValueError
        Input error, Input arrays hWave is not 1D
    ValueError
        Input error, Input array tWave is not 1D
    ValueError
        Input arrays hWave and tWave have different length

    Example
    -------
    >>> import numpy as np
    >>> hWave = np.asrange([2,4,3])
    >>> tWave = np.asrange([6,7,8])
    >>> # Then we get:
    >>> [hWaveSorted, tWaveSorted] = sort_wave_params(hWave,tWave)
    >>> #hWaveSorted = np.asrange([4,3,2])
    >>> #tWaveSorted = np.asrange([7,8,6])  (and not: tWaveSorted = [8 7 6])

    """
    # --- ensure input is of type ndarray
    hWave = core_engine.convert_to_array_type(hWave)
    tWave = core_engine.convert_to_array_type(tWave)

    # Check on input arguments
    # --- Check whether the size of the input arguments is identical
    is1d_array = core_engine.is1darray(hWave)
    if not is1d_array:
        raise ValueError("sort_wave_params: Input error, Input arrays hWave is not 1D")

    is1d_array = core_engine.is1darray(tWave)
    if not is1d_array:
        raise ValueError("sort_wave_params: Input error, Input array tWave is not 1D")

    # --- Check whether number of elements in input arguments is identical
    isequal_length = len(hWave) == len(tWave)
    if not isequal_length:
        raise ValueError(
            "InputError", "Input arrays hWave and tWave have different length"
        )

    # --- Number of waves
    nWave = len(hWave)

    # --- If there are no waves, return to calling function
    hWaveSorted = np.empty((0, 0))
    tWaveSorted = np.empty((0, 0))
    if nWave < 1:
        return hWaveSorted, tWaveSorted

    # Computational core
    # --- Sort the wave height in descending order
    iSorted = np.argsort(hWave)[
        ::-1
    ]  # term [::-1] reverses the array, meaning descending order.
    hWaveSorted = hWave[iSorted]
    # --- Sort the wave periods
    tWaveSorted = tWave[iSorted]
    #
    return hWaveSorted, tWaveSorted


def determine_zero_crossing(
    t: NDArray[float64],
    xTime: NDArray[float64],
    typeCross: str = "down",
) -> tuple[int, NDArray[float64]]:
    """Determines zero-crossings (number and positions) of signal.

    This function determines the zero-crossings of a given time signal xTime = xTime(t). Here, t stands for time, and
    xTime for a given signal (for example, time series of measured surface elevation). The type of zero-crossing can
    be either up-crossing or down-crossing.

    Parameters
    ----------
    t : NDArray[float64]
        1D array containing time axis. The numbers in the array t must be increasing and uniformly spaced
        (uniform time step) [s]
    xTime : NDArray[float64]
        1D array containing signal values, i.e. the time series of the signal. The value xxTime(i) must be the signal
        value at time t(i). Usually water surface elevation [m]
    typeCross : str, optional
        Search for up- or down-crossings, by default "down"

    Returns
    -------
    tuple[int, NDArray[float64]]
        nWave : int
            Number of waves in the signal, where one wave corresponds to two successive zero-crossings. Wave i starts
            at time tCross(i), and end at time tCross(i+1) [-]
        tCross : NDArray[float64]
            1D array of length (nWave+1), containing the time of all zero-crossings. The time of the zero-crossings is
            determined by linear interpolation. Note that in case of no zero-crossing, the array tCross is empty. Note
            that in case of one zero-crossing, the number of waves is zero. [s]

    Raises
    ------
    ValueError
        Input error: Input arrays t is not 1D
    ValueError
        Input error: Input arrays x is not 1D
    ValueError
        Input error: Input arrays t and x have different length
    ValueError
        Input error: Wrong input argument for typeCross

    """
    # --- ensure input is of type ndarray
    t = core_engine.convert_to_array_type(t)
    xTime = core_engine.convert_to_array_type(xTime)

    # Perform checks on the input arguments
    is1d_array = core_engine.is1darray(t)
    if not is1d_array:
        raise ValueError("sort_wave_params: Input error: Input arrays t is not 1D")

    is1d_array = core_engine.is1darray(xTime)
    if not is1d_array:
        raise ValueError("sort_wave_params: Input error: Input arrays x is not 1D")

    # --- Check whether the length of arrays t and x the same
    isequal_length = len(t) == len(xTime)
    if not isequal_length:
        raise ValueError(
            "sort_wave_params: Input error: Input arrays t and x have different length"
        )

    # Computational core
    nWave = -1
    tCross = np.empty((0, 0))
    dt = t[1] - t[0]
    nTime = len(t)

    # --- Loop over all time levels, to determine zero-crossings
    if typeCross.lower() == "up":
        # --- Upcrossings
        for iTime in np.arange(0, nTime - 1):
            if xTime[iTime] <= 0 and xTime[iTime + 1] > 0:
                # --- Add 1 to counter number of zero-crossings
                nWave = nWave + 1

                # --- Determine time level of zero-crossing by means of linear
                #     interpolation
                tCrossing = t[iTime] + dt * xTime[iTime] / (
                    xTime[iTime] - xTime[iTime + 1]
                )
                tCross = np.append(tCross, tCrossing)

    elif typeCross.lower() == "down":
        # --- Downcrossings
        for iTime in np.arange(0, nTime - 1):
            if xTime[iTime] >= 0 and xTime[iTime + 1] < 0:
                # --- Add 1 to counter number of zero-crossings
                nWave = nWave + 1

                # --- Determine time level of zero-crossing by means of linear
                #     interpolation
                tCrossing = t[iTime] + dt * xTime[iTime] / (
                    xTime[iTime] - xTime[iTime + 1]
                )
                tCross = np.append(tCross, tCrossing)
    else:
        raise ValueError(
            "sort_wave_params: Input error: Wrong input argument for typeCross "
        )

    # --- If no zero-crossings are found, put number of waves to zero (the
    #     value -1 may be considered inappropriate)
    if nWave == -1:
        nWave = 0

    return nWave, tCross


def highest_waves_params(
    hWaveSorted: NDArray[float64],
    tWaveSorted: NDArray[float64],
    fracP: float,
) -> tuple[float, float]:
    """Computes wave parameters of selection largest waves

    This function computes the wave height hFracP and wave period tFracP by taking the average of the fraction fracP
    of the highest waves. When fracP = 1/3, then hFracP is equal to the significant wave height and tFracP is equal
    to the significant wave period.

    Parameters
    ----------
    hWaveSorted : NDArray[float64]
        1D array containing wave heights of the individual waves, sorted in descending order [m]
    tWaveSorted : NDArray[float64]
        1D array containing wave periods of the individual waves, using the same sorting (re-arranging) as
            applied to hWave [s]
    fracP : float
        fraction. Should be between 0 and 1 [-]

    Returns
    -------
    tuple[float, float]
        hFracP : float
            average of the wave heights of the highest fracP waves [m]
        tFracP : float
            average of the wave periods of the highest fracP waves [s]

    Raises
    ------
    ValueError
        Input error: Input arrays hWaveSorted is not 1D
    ValueError
        Input error: Input arrays tWaveSorted is not 1D
    ValueError
        Input arrays tWaveSorted and hWaveSorted have different length
    ValueError
        Input error: Input parameter fracP should be between 0 and 1
    ValueError
        Input error:Input array hWaveSorted is not correctly sorted

    Example
    -------
    >>> import numpy as np
    >>> hWaveSorted = [7.1, 6.0, 5.8, 4.4, 3.1, 2.2, 1.4]
    >>> tWaveSorted = [5.1, 3.2, 3.7, 0.0, 2.0, 3.0, 1.1]
    >>> [Hsig,Tsig] = highest_waves_params(hWaveSorted,tWaveSorted,1/3)

    """
    # --- ensure input is of type ndarray
    hWaveSorted = core_engine.convert_to_array_type(hWaveSorted)
    tWaveSorted = core_engine.convert_to_array_type(tWaveSorted)

    # Perform checks on the input arguments
    is1d_array = core_engine.is1darray(hWaveSorted)
    if not is1d_array:
        raise ValueError(
            "highest_waves_params: Input error: Input arrays hWaveSorted is not 1D"
        )

    is1d_array = core_engine.is1darray(tWaveSorted)
    if not is1d_array:
        raise ValueError(
            "highest_waves_params: Input error: Input arrays tWaveSorted is not 1D"
        )

    # --- Check whether the length of arrays t and x the same
    isequal_length = len(tWaveSorted) == len(hWaveSorted)
    if not isequal_length:
        raise ValueError(
            "highest_waves_params: Input error: Input arrays tWaveSorted and hWaveSorted have different length"
        )

    # --- Input parameter fracP should be between 0 and 1
    if fracP < 0 or fracP > 1:
        raise ValueError(
            "highest_waves_params: Input error: Input parameter fracP should be between 0 and 1 "
        )

    # --- If there are no waves, return to calling function
    hFracP = 0.0
    tFracP = 0.0

    # --- Number of waves
    nWave = len(hWaveSorted)
    if nWave < 1:
        return hFracP, tFracP

    # --- Check whether hWaveSorted is indeed sorted
    issorted = (hWaveSorted == np.sort(hWaveSorted)[::-1]).all()
    if not issorted:
        raise ValueError(
            "highest_waves_params: Input error:Input array hWaveSorted is not correctly sorted "
        )

    # Computational core
    # --- Number of waves in the fraction that is to be considered
    nWaveP = int(np.floor(nWave * fracP))
    if nWaveP > 0:
        hFracP = float(np.mean(hWaveSorted[0:nWaveP]))
        tFracP = float(np.mean(tWaveSorted[0:nWaveP]))

    return hFracP, tFracP


def exceedance_wave_height(hWaveSorted: NDArray[float64], excPerc: float) -> float:
    """Computes wave height with given exceedance probability

    This function computes the wave height hExcPerc with given exceedance probability percentage excPerc.

    Parameters
    ----------
    hWaveSorted : NDArray[float64]
        1D array containing wave heights of the individual waves, sorted in descending order [m]
    excPerc : float
        exceedance probability percentage. excPerc = 2 means an exceedance percentage of 2%. The value of excPerc
        should not exceed 100, or be smaller than 0 [%]

    Returns
    -------
    float
        hExcPerc : float
            wave height with given exceedance probability [m]

    Raises
    ------
    ValueError
        Input error: Input arrays hWaveSorted is not 1D
    ValueError
        Input error: Input parameter excPerc should be between 0 and 100
    ValueError
        Input error:Input array hWaveSorted is not correctly sorted

    Example
    -------
    >>> import numpy as np
    >>> hWaveSorted = [7.1, 6.0, 5.8, 4.4, 3.1, 2.2, 1.4]
    >>> [hExcPerc_33perc] = exceedance_wave_height(hWaveSorted,33)

    """
    hWaveSorted = core_engine.convert_to_array_type(hWaveSorted)

    # Perform checks on the input arguments
    is1d_array = core_engine.is1darray(hWaveSorted)
    if not is1d_array:
        raise ValueError(
            "exceedance_wave_height: Input error: Input arrays hWaveSorted is not 1D"
        )

    # Check on input arguments
    # --- Input parameter excPerc should not exceed 100 (100%) or be smaller than 0
    isvalid_range = excPerc > 100 or excPerc < 0
    if isvalid_range:
        raise ValueError(
            "exceedance_wave_height: Input error: Input parameter excPerc should be between 0 and 100 "
        )

    # --- If there are no waves, return to calling function
    hExcPerc = 0.0
    # --- Number of waves
    nWave = len(hWaveSorted)
    if nWave < 1:
        return hExcPerc

    # --- Check whether hWaveSorted is indeed sorted
    issorted = (hWaveSorted == np.sort(hWaveSorted)[::-1]).all()
    if not issorted:
        raise ValueError(
            "exceedance_wave_height: Input error:Input array hWaveSorted is not correctly sorted "
        )

    # Computational core
    # --- Index of the wave corresponding to the given exceedance probability percentage
    iWaveP = int(np.floor(nWave * excPerc / 100.0)) - 1

    # --- Compute wave height with exceedance probability
    if not (iWaveP < 0):
        hExcPerc = float(hWaveSorted[iWaveP])

    return hExcPerc


def determine_params_individual_waves(
    tCross: NDArray[float64],
    t: NDArray[float64],
    xTime: NDArray[float64],
) -> tuple[
    NDArray[float64],
    NDArray[float64],
    NDArray[float64],
    NDArray[float64],
    NDArray[float64],
    NDArray[float64],
]:
    """Determines parameters per individual wave

    This function determines several wave properties (wave period, wave height, crest amplitude, trough amplitude, and
    the time at which the maximum crest and trough amplitudes occur) of all individual waves in a wave train. The wave
    train is given by time series xTime = xTime(t), and the zero-crossings occur at time levels given in tCross

    Parameters
    ----------
    tCross : NDArray[float64]
        1D array of length (nWave+1), containing the time of all zero-crossings. The time of the zero-crossings is
        determined by linear interpolation. Note that in case of no zero-crossing, the array tCross is empty. Note
        that in case of one zero-crossing, the number of waves is zero. [s]
    t : NDArray[float64]
        1D array containing time axis. The numbers in the array t must be increasing and uniformly spaced
        (uniform time step) [s]
    xTime : NDArray[float64]
        1D array containing signal values, i.e. the time series of the signal. The value xTime(i) must be the signal
        value at time t(i). Usually water surface elevation [m]

    Returns
    -------
    tuple[ NDArray[float64], NDArray[float64], NDArray[float64], NDArray[float64], NDArray[float64], NDArray[float64], ]
        tWave : NDArray[float64]
            1D array containing the periods of the individual waves [s]
        hWave : NDArray[float64]
            1D array containing the wave heights of the individual waves [m]
        aCrest : NDArray[float64]
            1D array containing the maximum amplitude of the crests of the individual waves [m]
        aTrough : NDArray[float64]
            1D array containing the maximum amplitude of the troughs of the individual waves [m]
        tCrest : NDArray[float64]
            1D array containing the time at which maximum crest amplitude of the individual waves occurs [s]
        tTrough : NDArray[float64]
            1D array containing the time at which maximum trough amplitude of the individual waves occurs [s]

    Notes
    -----
    * All these arrays have a length equal to nWave, which is the number of waves in the wave train
    * The values of aTrough are always smaller than zero
    * hWave = aCrest - aTrough

    Raises
    ------
    ValueError
        Input error: Input arrays tCross is not 1D
    ValueError
        Input error: Input arrays t is not 1D
    ValueError
        Input error: Input arrays x is not 1D
    ValueError
        Length of input arrays t and x not identical

    """
    # --- ensure input is of type ndarray
    tCross = core_engine.convert_to_array_type(tCross)
    t = core_engine.convert_to_array_type(t)
    xTime = core_engine.convert_to_array_type(xTime)

    # Perform checks on input arguments
    # --- In the possible situation of no zero-crossings, the array tCross is
    #     empty. In this situation, make output variables empty arrays and
    #     return
    tWave = np.empty((0, 0))
    hWave = np.empty((0, 0))
    aCrest = np.empty((0, 0))
    aTrough = np.empty((0, 0))
    tCrest = np.empty((0, 0))
    tTrough = np.empty((0, 0))
    if tCross.size == 0:
        return hWave, tWave, aCrest, aTrough, tCrest, tTrough

    # Perform checks on the input arguments
    is1d_array = core_engine.is1darray(tCross)
    if not is1d_array:
        raise ValueError(
            "determine_params_individual_waves: Input error: Input arrays tCross is not 1D"
        )

    is1d_array = core_engine.is1darray(t)
    if not is1d_array:
        raise ValueError(
            "determine_params_individual_waves: Input error: Input arrays t is not 1D"
        )

    is1d_array = core_engine.is1darray(xTime)
    if not is1d_array:
        raise ValueError(
            "determine_params_individual_waves: Input error: Input arrays x is not 1D"
        )

    # --- Check whether the length of arrays t and x the same
    isequal_length = len(t) == len(xTime)
    if not isequal_length:
        raise ValueError(
            "determine_params_individual_waves: Length of input arrays t and x not identical "
        )

    # Computational core
    # --- Determine number of waves
    nWave = len(tCross) - 1

    # --- Initialize output arrays (could have used append also, without initializing array)
    tWave = np.zeros((1, nWave), dtype=float)[0]
    hWave = np.zeros((1, nWave), dtype=float)[0]
    aCrest = np.zeros((1, nWave), dtype=float)[0]
    aTrough = np.zeros((1, nWave), dtype=float)[0]
    tCrest = np.zeros((1, nWave), dtype=float)[0]
    tTrough = np.zeros((1, nWave), dtype=float)[0]

    # --- Do a loop over all waves
    for iWave in np.arange(0, nWave):
        # --- Determine initial and end time of wave, and corresponding
        #     positions in time-array
        tIni = tCross[iWave]  # Initial time
        tEnd = tCross[iWave + 1]  # End time
        iIni = np.where(t >= tIni)[0][0]
        iEnd = np.where(t <= tEnd)[0][-1]

        # --- Find position where maximum and minimum occurs, for the
        #     individual wave
        #     Argument 'first' is included, to account for the possible
        #     situation of two extrema with the same function
        #     value in the given wave
        tIwave = t[iIni : iEnd + 1]
        xIwave = xTime[iIni : iEnd + 1]
        iMax = np.where(xIwave == max(xIwave))[0][0]
        iMin = np.where(xIwave == min(xIwave))[0][0]

        # --- Fill the output arrays
        tWave[iWave] = tEnd - tIni
        aCrest[iWave] = xIwave[iMax]
        aTrough[iWave] = xIwave[iMin]
        tCrest[iWave] = tIwave[iMax]
        tTrough[iWave] = tIwave[iMin]
        hWave[iWave] = aCrest[iWave] - aTrough[iWave]

    return hWave, tWave, aCrest, aTrough, tCrest, tTrough



# SPDX-License-Identifier: GPL-3.0-or-later
import numpy as np
from numpy import float64
from numpy.typing import NDArray

import deltares_wave_toolbox.cores.core_engine as core_engine


def disper(w: NDArray[float64], h: float, g: float = 9.81):
    """DISPER  Solves the wave number from the linear dispersion relation

    The linear dispersion relation reads:

    w^2 = g * k * tanh( k * h )

    where w is the radial frequency, g the gravitation acceleration
    constant, k the wave number, and h the water depth. The function
    DISPER solves for the wave number k, given the radial frequency w,
    the water depth h and the gravitational acceleration constant g.

    water depth, units meter
         Three possible options:
         (i)   0 < h < infinity: h is water depth, and nonlinear dispersion
               relation is solved using given value for h
         (ii)  h < 0. The shallow water dispersion relation is solved, with
               the actual water depth being equal to |h|
         (iii) h = inf. The deep water dispersion relation is solved.

    This function is originally written by G. Klopman, Delft Hydraulics, 6
    Dec 1994. It is stated that the relative error in k*h < 2.5e-16 for
    all k*h

    Parameters
    ----------
    w : NDArray[float64]
        radial frequency ( = 2 * pi / wave period), unit: rad/s.
    h : float
        water depth, unit: meters
    g : float, optional
        representing gravitational constant, by default 9.81

    Returns
    -------
    k : NDArray[float64]
        wave number
    """

    # check if radial frequency is single value or of type array, if of type single value
    # convert to array to ensure this function can handle single values as well as arrays.
    w = core_engine.convert_to_array_type(w)

    # if not (isinstance(w,np.ndarray)):
    #   w = w*np.ones(1)

    if np.isinf(h):
        # --- Deep water dispersion relation
        k = w**2 / g  # element wise multiplication.
    elif h < 0:
        # --- Shallow water dispersion relation, with depth equal to |h|
        k = w / np.sqrt(g * abs(h))
    else:
        # --- Standard (nonlinear) dispersion relation for depths 0 < h < inf
        w2 = (w**2) * h / g

        ielem = np.nonzero(w2 < 1.0e-8)
        w2[ielem] = 1e-8
        q = w2 / (1 - np.exp(-(w2 ** (5 / 4)))) ** (2 / 5)

        idxs = [1, 2]
        for j in idxs:
            thq = np.tanh(q)
            thq2 = 1 - thq**2
            a = (1 - q * thq) * thq2
            b = thq + q * thq2
            c = q * thq - w2
            arg = np.zeros(np.size(q))
            iq = np.where(a != 0)[0]
            arg[iq] = (b[iq] ** 2) - 4 * a[iq] * c[iq]
            arg[iq] = (-b[iq] + np.sqrt(arg[iq])) / (2 * a[iq])
            iq = np.nonzero(abs(a * c) < 1.0e-8 * (b**2))
            arg[iq] = -c[iq] / b[iq]
            q = q + arg

    k = np.sign(w) * q / h

    ik = np.isnan(k)
    k[ik] = np.zeros(np.size(k[ik]))

    return k



# SPDX-License-Identifier: GPL-3.0-or-later
from .cores import *  # noqa
from .series import Series, WaveHeights  # noqa
from .spectrum import Spectrum  # noqa

__version__ = "1.1.0"

import os

# set stylesheet
import matplotlib.pyplot as plt

plt.style.use(
    os.path.join(os.path.dirname(os.path.abspath(__file__)), "Style.mplstyle")
)



# SPDX-License-Identifier: GPL-3.0-or-later
import matplotlib.pyplot as plt
import numpy as np
from matplotlib import figure
from numpy import complex128, float64
from numpy.typing import NDArray
from scipy.signal import hilbert
from scipy.stats import rayleigh

import deltares_wave_toolbox.cores.core_engine as core_engine
import deltares_wave_toolbox.cores.core_spectral as core_spectral
import deltares_wave_toolbox.cores.core_time as core_time
import deltares_wave_toolbox.cores.core_wavefunctions as core_wavefunctions
import deltares_wave_toolbox.spectrum as spectrum


class WaveHeights:
    """The WaveHeights class

    Parameters
    ----------
    hwave : NDArray[float64]
        1D array containing the wave heights of the individual waves [m]
    twave : NDArray[float64]
        1D array containing the periods of the individual waves [s]

    """

    def __init__(self, hwave: NDArray[float64], twave: NDArray[float64]) -> None:
        hwave, _ = core_engine.convert_to_vector(hwave)
        twave, _ = core_engine.convert_to_vector(twave)

        self.hwave = hwave
        self.nwave = len(hwave)
        self.twave = twave

    def __str__(self) -> str:
        return f"Series with {self.nwave} elements"

    def __repr__(self) -> str:
        return f"{type(self).__name__} (series  nt = {self.nwave})"

    def sort(self) -> None:
        """Sort wave heights and wave periods

        Sorts the wave height and wave period.
        The sorting is done such that in hWaveSorted the wave heights of hWave
        are sorted in descending order. This same sorting is applied to
        tWave.

        """
        self.hwave, self.twave = core_time.sort_wave_params(self.hwave, self.twave)

    def get_Hrms(self) -> float:
        """Compute root-mean-squared wave height

        Returns
        -------
        float
            Hrms

        """
        return np.sqrt(np.mean(self.hwave**2))

    def get_Hmax(self) -> float64:
        """Return maximum wave height

        Returns
        -------
        float64
            Hmax

        """
        return np.max(self.hwave)

    def get_Hs(self) -> tuple[float, float]:
        """Compute significant wave height & associated period

        Hs (significant wave height) is the average of the highest 1/3 of the waves

        Returns
        -------
        tuple[float, float]
            Hs, Ts

        """
        return self.highest_waves(1 / 3)

    def get_H2p_Rayleigh(self) -> float:
        """Compute 2% exceedance wave height assuming theoretical Rayleigh distribution

        Returns
        -------
        float
            H2p

        """
        return self.get_Hs()[0] * rayleigh.ppf(0.98, scale=1 / np.sqrt(2)) / np.sqrt(2)

    def get_exceedance_waveheight(self, excPerc: float) -> float:
        """Computes wave height with given exceedance probability

        This function computes the wave height hExcPerc with given exceedance probability percentage excPerc.

        Parameters
        ----------
        excPerc : float
            exceedance probability percentage. excPerc = 2 means an exceedance percentage of 2%. The value of excPerc
            should not exceed 100, or be smaller than 0 [%]

        Returns
        -------
        float
            wave height with given exceedance probability [m]

        """
        self.hwave, self.twave = core_time.sort_wave_params(self.hwave, self.twave)
        return core_time.exceedance_wave_height(hWaveSorted=self.hwave, excPerc=excPerc)

    def highest_waves(self, fracP: float) -> tuple[float, float]:
        """Computes wave parameters of selection largest waves

        This function computes the wave height hFracP and wave period tFracP by taking the average of the fraction
        fracP of the highest waves. When fracP = 1/3, then hFracP is equal to the significant wave height and tFracP
        is equal to the significant wave period.

        Parameters
        ----------
        fracP : float
            fraction. Should be between 0 and 1 [-]

        Returns
        -------
        tuple[float, float]
            hFracP : float
                average of the wave heights of the highest fracP waves [m]
            tFracP : float
                average of the wave periods of the highest fracP waves [s]

        """
        self.hwave, self.twave = core_time.sort_wave_params(self.hwave, self.twave)
        hFracP, tFracP = core_time.highest_waves_params(
            hWaveSorted=self.hwave, tWaveSorted=self.twave, fracP=fracP
        )
        return hFracP, tFracP

    def plot_exceedance_waveheight(self, savepath: str = "") -> figure.Figure:
        """Plot exceedances of wave heights

        Parameters
        ----------
        savepath : str, optional
            path to save figure, by default ""

        Returns
        -------
        figure.Figure
            figure object

        """
        self.hwave, self.twave = core_time.sort_wave_params(self.hwave, self.twave)
        fig = plt.figure()
        plt.plot(
            self.hwave, np.linspace(0, self.nwave, self.nwave, self.nwave) / self.nwave
        )

        plt.yscale("log")
        plt.grid("on")
        plt.xlabel("Wave height [$m$]")
        plt.ylabel("P exceedance ")

        if savepath != "":
            plt.savefig(savepath)
        return fig

    def plot_exceedance_waveheight_Rayleigh(
        self,
        savepath: str = "",
        normalize: bool = False,
        plot_BG: bool = False,
        water_depth: float = -1.0,
        cota_slope: float = 250.0,
        hm0: float = -1.0,
    ) -> figure.Figure:
        """Plot exceedances of wave heights compared to Rayleigh distribution

        Parameters
        ----------
        savepath : str, optional
            path to save figure, by default ""
        normalize : bool, optional
            normalize wave heights with significant wave height (Hs), by default False
        plot_BG : bool, optional
            include theoretical Battjes & Groenendijk (2000) distribution in plot, by default False
        water_depth : float, optional
            water depth needed for Battjes & Groenendijk (2000), by default -1.0 [m]
        cota_slope : float, optional
            foreshore slope needed for Battjes & Groenendijk (2000), by default 250.0 [-]
        hm0 : float, optional
            spectral wave height needed for Battjes & Groenendijk (2000), by default -1.0 [m]

        Returns
        -------
        figure.Figure
            figure object

        Raises
        ------
        ValueError
            Raised when plot_BG is True and Hm0 is not provided

        """
        if normalize:
            # Normalize with significant wave height
            H_normalize = self.get_Hs()[0]
            y_label = r"Normalized wave height $\frac{H_{i}}{H_{s}}$ [-]"
        else:
            # no normalization
            H_normalize = 1
            y_label = r"Wave height $H_{i}$ [$m$]"

        Rayleigh_x = np.sqrt(-np.log(np.arange(1, self.nwave + 1) / self.nwave))
        H2p_Rayleigh = self.get_H2p_Rayleigh()
        Rayleigh_theoretical_dist = H2p_Rayleigh * np.sqrt(
            np.log(1 - np.arange(self.nwave, 0, -1) / (self.nwave + 1)) / np.log(0.02)
        )

        self.hwave, self.twave = core_time.sort_wave_params(self.hwave, self.twave)
        fig = plt.figure()

        plt.plot(
            Rayleigh_x,
            Rayleigh_theoretical_dist / H_normalize,
            label="Theoretical Rayleigh distribution",
        )

        if plot_BG:
            if hm0 == -1.0:
                raise ValueError(
                    "Please provide Hm0 when using Battjes & Groenendijk distribution"
                )
            else:
                (
                    hwave_BG,
                    Pexceedance_BG,
                ) = core_wavefunctions.compute_BattjesGroenendijk_wave_height_distribution(
                    hm0, self.nwave, water_depth, cota_slope=cota_slope
                )

                plt.plot(
                    np.sqrt(-np.log(Pexceedance_BG)),
                    hwave_BG / H_normalize,
                    label="Battjes & Groenendijk (2000) distribution",
                )

        plt.plot(Rayleigh_x, self.hwave / H_normalize, label="Wave height distribution")

        plt.grid("on")
        plt.xlabel(r"$P_{exceedance}$ [$\%$]")
        plt.ylabel(y_label)
        plt.legend()

        xtick_positions = np.array([100, 50, 20, 10, 5, 2, 1, 0.1, 0.01, 0.001])
        ax = plt.gca()
        ax.set_xticks(
            rayleigh.ppf(1 - xtick_positions / 100, scale=1 / np.sqrt(2)),
            labels=xtick_positions,
        )

        if savepath != "":
            plt.savefig(savepath)
        return fig

    def plot_hist_waveheight(self, savepath: str = "") -> figure.Figure:
        """Plot Histogram of wave heights

        Parameters
        ----------
        savepath : str, optional
            path to save figure, by default ""

        Returns
        -------
        figure.Figure
            figure object

        """
        fig = plt.figure()
        plt.hist(self.hwave, label="Distribution")

        plt.grid("on")
        plt.xlabel("H [$m$]")
        plt.ylabel("P ")
        plt.legend()

        if savepath != "":
            plt.savefig(savepath)
        return fig


class Series(WaveHeights):
    """The wave Series class

    Contains a time series (typically) of water level elevations. Inherits from WaveHeights class.

    Parameters
    ----------
    time : NDArray[float64]
        1D array containing time axis. The numbers in the array t must be increasing and uniformly spaced
        (uniform time step) [s]
    xTime : NDArray[float64]
        1D array containing signal values, i.e. the time series of the signal. The value xTime(i) must be the signal
        value at time t(i). Usually water surface elevation [m]

    """

    def __init__(self, time: NDArray[float64], xTime: NDArray[float64]) -> None:
        time, tSize = core_engine.convert_to_vector(time)
        xTime, xSize = core_engine.convert_to_vector(xTime)

        assert tSize[0] == xSize[0], "Input error: array sizes differ in dimension"
        assert np.var(np.diff(time)) < np.mean(
            np.diff(time) / 100
        ), "Input error: time step is not equidistant"

        self.time = time
        self.xTime = xTime
        self.nt = len(time)
        self.dt = np.mean(np.diff(self.time))

        [
            hWave,
            tWave,
            _,
            _,
            _,
            _,
            _,
        ] = self._determine_individual_waves()
        super().__init__(hWave, tWave)

    def __str__(self) -> str:
        return f"Series with {self.nt} elements"

    def __repr__(self) -> str:
        return f"{type(self).__name__} (series  nt = {self.nt})"

    def get_crossing(self, typeCross: str = "down") -> tuple[int, NDArray[float64]]:
        """Get zero crossings form time series

        Determine either the zero up- or down-crossings of the time series.

        Parameters
        ----------
        typeCross : str, optional
            Search for up- or down-crossings, by default "down"

        Returns
        -------
        tuple[int, NDArray[float64]]
            nWave : int
                Number of waves in the signal, where one wave corresponds to two successive zero-crossings. Wave i
                starts at time tCross(i), and end at time tCross(i+1) [-]
            tCross : NDArray[float64]
                1D array of length (nWave+1), containing the time of all zero-crossings. The time of the
                zero-crossings is determined by linear interpolation. Note that in case of no zero-crossing, the
                array tCross is empty. Note that in case of one zero-crossing, the number of waves is zero. [s]

        """
        nWave, tCross = core_time.determine_zero_crossing(
            t=self.time, xTime=self.xTime, typeCross=typeCross
        )
        return nWave, tCross

    def get_skewness(self) -> tuple[float, float]:
        """Compute skewness

        Sk is skewness of the signal

        Returns
        -------
        float
            Sk

        """
        Sk = np.mean((self.xTime - np.mean(self.xTime)) ** 3) / np.mean(
            (self.xTime - np.mean(self.xTime)) ** 2
        ) ** (1.5)
        return Sk

    def get_asymmetry(self) -> tuple[float, float]:
        """Compute asymmetry

        As is asymmetry of the signal

        Returns
        -------
        float
            As

        """
        H = np.imag(hilbert(self.xTime))
        As = np.mean(H**3) / np.mean((self.xTime - np.mean(self.xTime)) ** 2) ** (1.5)
        return As

    def get_spectrum(
        self,
        nperseg: int = 256,
        noverlap: int = 0,
        nfft: int = 0,
        windows_type: str = "hann",
        dfDesired: float = 0.01,
        use_dfDesired: bool = True,
    ) -> spectrum.Spectrum:
        """create spectrum

        Parameters
        ----------
        nperseg : int, optional
            Length of each segment, by default None
        noverlap : int, optional
            number of points in overlap, by default None
        nfft : int, optional
            length of fft, by default None
        window_type : str, optional
            window type, by default "hann"

        Returns
        -------
        spectrum.Spectrum
            Spectrum in spectrum object

        """

        if use_dfDesired:
            nperseg = np.round((1 / self.dt) / dfDesired).astype(int)
            noverlap = np.round(nperseg / 2).astype(int)

        [f, S] = core_spectral.compute_spectrum_welch_wrapper(
            self.xTime,
            dt=self.dt,
            nperseg=nperseg,
            noverlap=noverlap,
            nfft=nfft,
            window_type=windows_type,
        )
        return spectrum.Spectrum(f, S)

    def get_spectrum_raw(self, dfDesired: float = 0.01) -> spectrum.Spectrum:
        """create spectrum with a desired frequency, but without applying segments

        Parameters
        ----------
        dfDesired : float, optional
            desired frequency spacing in Hertz on which the wave spectrum must be computed.
            If this parameter is omitted, then dfDesired = f(1) - f(0), by default 0.01 [Hz]

        Returns
        -------
        spectrum.Spectrum
            Spectrum in spectrum object

        """
        [f, S] = core_spectral.compute_spectrum_time_series(
            self.time, self.xTime, dfDesired
        )
        return spectrum.Spectrum(f, S)

    def max(self) -> float:
        return np.max(self.xTime)

    def min(self) -> float:
        return np.min(self.xTime)

    def mean(self) -> float:
        return np.mean(self.xTime)

    def var(self) -> float:
        return np.var(self.xTime)

    def get_bandpassfilter(self, fmin: float, fmax: float):
        """get bandpassfiltered signal

        Parameters
        ----------
        fmin : lower frequency for bandpassfilter
        fmax : higher frequency for bandpassfilter
        Returns
        -------
        series.Series
            Series in series object

        """
        xTimeFiltered = core_spectral.bandpassfilter(self.time, self.xTime, fmin, fmax)
        xTimeFiltered = Series(self.time, xTimeFiltered)
        return xTimeFiltered

    def get_fourier_comp(
        self,
    ) -> tuple[NDArray[float64], NDArray[complex128], bool]:
        """get Fourier components from series

        Returns
        -------
        tuple[NDArray[float64], NDArray[complex128], bool]
            f : NDArray[float64]
                1D array containing frequency values, for folded Fourier transform. The frequency axis runs from 0 to
                the Nyquist frequency. The number of elements in array f is close to half the number of elements in
                array xTime. To be precise, length(f) = floor(nT/2) + 1, with nT the number of elements in array
                xTime [Hz]
            xFreq : NDArray[complex128]
                1D array (complex!) containing the folded Fourier coefficients of xTime. The value xFreq(i) must be the
                Fourier coefficient at frequency f(i). The number of elements in f and xFreq are the same.
            isOdd : bool
                logical indicating whether nT, the number of time points in xTime, is even (isOdd=False) or odd
                (isOdd=True)

        """
        f, xFreq, isOdd = core_spectral.time2freq_nyquist(self.time, self.xTime)
        return f, xFreq, isOdd

    def _determine_individual_waves(
        self, typeCross: str = "down", fcutoff=0, minimum_wave=0
    ) -> tuple[
        NDArray[float64],
        NDArray[float64],
        NDArray[float64],
        NDArray[float64],
        NDArray[float64],
        NDArray[float64],
    ]:
        """determine individual waves in series

        Parameters
        ----------
        typeCross : str, optional
            Search for up- or down-crossings, by default "down"

        Returns
        -------
        tuple[ NDArray[float64], NDArray[float64], NDArray[float64], NDArray[float64], NDArray[float64],
        NDArray[float64], ]
            tWave : NDArray[float64]
                1D array containing the periods of the individual waves [s]
            hWave : NDArray[float64]
                1D array containing the wave heights of the individual waves [m]
            aCrest : NDArray[float64]
                1D array containing the maximum amplitude of the crests of the individual waves [m]
            aTrough : NDArray[float64]
                1D array containing the maximum amplitude of the troughs of the individual waves [m]
            tCrest : NDArray[float64]
                1D array containing the time at which maximum crest amplitude of the individual waves occurs [s]
            tTrough : NDArray[float64]
                1D array containing the time at which maximum trough amplitude of the individual waves occurs [s]

        Notes
        -----
        * All these arrays have a length equal to nWave, which is the number of waves in the wave train
        * The values of aTrough are always smaller than zero
        * hWave = aCrest - aTrough

        """
        if fcutoff != 0:
            xtime_low = core_spectral.bandpassfilter(
                t=self.time, xTime=self.xTime, flow=0, fhigh=fcutoff
            )
        else:
            xtime_low = 0
        _, tCross = core_time.determine_zero_crossing(
            t=self.time, xTime=self.xTime - xtime_low, typeCross=typeCross
        )
        (
            hWave,
            tWave,
            aCrest,
            aTrough,
            tCrest,
            tTrough,
        ) = core_time.determine_params_individual_waves(
            tCross=tCross, t=self.time, xTime=self.xTime
        )
        # remove waves lower than minimum_wave
        remove_index = np.where(hWave < minimum_wave)[0]
        self.tCross = np.delete(tCross, remove_index)
        hWave = np.delete(hWave, remove_index)
        tWave = np.delete(tWave, remove_index)
        aCrest = np.delete(aCrest, remove_index)
        aTrough = np.delete(aTrough, remove_index)
        tCrest = np.delete(tCrest, remove_index)
        tTrough = np.delete(tTrough, remove_index)

        super().__init__(hWave, tWave)  # set values here instead of init
        return hWave, tWave, aCrest, aTrough, tCrest, tTrough, tCross

    def plot(self, savepath: str = "", plot_crossing: bool = False) -> figure.Figure:
        """Plot Series

        Parameters
        ----------
        savepath : str, optional
            path to save figure, by default ""
        plot_crossing : bool, optional
            plot zero crossings in figure, by default False

        Returns
        -------
        figure.Figure
            figure object

        """
        fig = plt.figure()
        plt.plot(self.time, self.xTime, label="series")
        if plot_crossing and hasattr(self, "tCross"):

            plt.plot(self.tCross, np.asarray(self.tCross) * 0, "ro", label="crossing")
        plt.grid("on")
        plt.xlabel("time [$s$]")
        plt.ylabel("z [$m$]")
        plt.legend()

        if savepath != "":
            plt.savefig(savepath)
        return fig
